#include<linux/init.h>
#include<linux/fs.h>
#include<linux/module.h>
#include<linux/buffer_head.h>
#include<linux/blkdev.h>
#include "ngfs.h"
#define DEBUG

void ngfs_read_inode(struct inode *);
int ngfs_readdir(struct file *, void *, filldir_t);
ssize_t ngfs_read( struct file *file, char *buf, size_t max, loff_t *offset );
ssize_t ngfs_write( struct file *file, char *buf, size_t max, loff_t *offset ); 
struct dentry * ngfs_lookup(struct inode *inode,struct dentry *dirent);


static struct inode_operations ngfs_inode_ops = {
				lookup : ngfs_lookup,
};

static struct file_operations ngfs_dir_ops = {
				read : generic_read_dir,
				readdir : ngfs_readdir,
};

static struct super_operations ngfs_ops = {
				read_inode : ngfs_read_inode,
};

int ngfs_readdir(struct file *fp,void *dirent, filldir_t filldir){
    int i;
    unsigned32 pos;
    struct buffer_head *bh;
    struct inode *inode=(fp->f_dentry->d_inode);
    struct ngfs_inode *in=(struct ngfs_inode *)&(inode->u.generic_ip);
    struct super_block *sb=inode->i_sb;
    struct ngfs_dentry *de;
    bh=sb_bread(sb,in->block_no); /* Read the blockno where data is stored */
again:
    pos=fp->f_pos;
    if(pos > in->size){
        printk(KERN_ALERT "NGFS : Position greater than file size\n");
        brelse(bh);
        return 0;
    }
    de=(struct ngfs_dentry *)(bh->b_data + pos % sb->s_blocksize);
    if(de->ino == 0){
        fp->f_pos+=sizeof(struct ngfs_dentry);
 /*       brelse(bh); */
        goto again;
    }
    else{
        filldir(dirent,de->name,de->len,pos,de->ino,DT_UNKNOWN);
    }
    fp->f_pos+=sizeof(struct ngfs_dentry);
    brelse(bh);
    printk(KERN_ALERT "NGFS : readdir called\n");
       return 1;
}

int ngfs_get_inode(struct inode *inode,char * name){
    int i;
			 struct buffer_head *bh;
			 struct ngfs_inode *in=(struct ngfs_inode *)&(inode->u.generic_ip);
			 struct super_block *sb=inode->i_sb;
			 struct ngfs_dentry *de;
    printk(KERN_ALERT "NGFS : getinode called\n");
			 printk(KERN_ALERT "NGFS : Reading data from block no. [%d] for name:inode [%s:%d]\n",in->block_no,name,in->ino);
    bh=sb_bread(sb,in->block_no);
    if(!bh)
      goto exit;
    de=(struct ngfs_dentry *)bh->b_data;
   for(i=0;i<(sb->s_blocksize)/sizeof(*de);i++){
    if(strcmp(de->name,""))
      printk(KERN_ALERT "Read dentry name [%s]\n",de->name);
    if(strcmp(name,de->name)==0){
      brelse(bh);
      return de->ino;
    }
    de++;
   }
exit:
    return 0;
}

struct dentry * ngfs_lookup(struct inode *inode,struct dentry *dirent){
			 int inode_num;
			 struct inode *ino=NULL;
       inode_num=0;
			 printk(KERN_ALERT "NGFS:Trying to lookup file \"%s\"\n",(char *)dirent->d_name.name);
			 printk(KERN_ALERT "NGFS: superblock at [0x%x]\n",inode->i_sb);
			 inode_num=ngfs_get_inode(inode,(char *)dirent->d_name.name);
			 if(inode_num!=0){
							 ino=iget(inode->i_sb,inode_num);
				    if(!ino){
								printk(KERN_ALERT "NGFS:Cannot read inode number [%d] \n",inode_num);
								return ERR_PTR(-EINVAL);

			 }
       printk(KERN_ALERT "NGFS : lookup called on file \"%s\"\n",dirent->d_name);
			 d_add(dirent,ino);
			 }
       return NULL;
}

ssize_t ngfs_read( struct file *file, char *buf, size_t max, loff_t *offset ){
       printk(KERN_ALERT "NGFS : read called\n");
       return 1;
}
ssize_t ngfs_write( struct file *file, char *buf, size_t max, loff_t *offset ){
       printk(KERN_ALERT "NGFS : write called\n");
       return 1;
}

/*
void dump_super_block(struct ngfs_super_block *sb){
				printk(KERN_ALERT "Dumping super block of NGFS :
				                   root = %d,
													 block_bmp = %d,
													 inode_bmp = %d,
													 block_size = %d,
													 blocks = %d,
													 s_magic = 0x%x,
													 magic = %s\n",sb->root,sb->block_bmp,sb->inode_bmp,sb->block_size,sb->blocks,sb->s_magic,
													 sb->magic);
}
*/

void ngfs_read_inode(struct inode *inode){
				struct buffer_head *bh;
				struct ngfs_inode *ngfs_ino;
			  if (!(bh = sb_bread(inode->i_sb, NGFS_INODE_BLOCK))) {
								printk(KERN_ALERT "Cannot read inode [%d] of NGFS \n",inode->i_ino);
								goto free;
				}
				ngfs_ino=(struct ngfs_inode *)((char *)bh->b_data+inode->i_ino*sizeof(*ngfs_ino));
				inode->i_mode=ngfs_ino->mode;
				inode->i_uid=ngfs_ino->uid;
				inode->i_gid=ngfs_ino->gid;
				inode->i_blocks=ngfs_ino->block_no;
				inode->i_nlink=ngfs_ino->links;
    inode->i_size=ngfs_ino->size;
    memcpy(&inode->u.generic_ip,ngfs_ino,sizeof(struct ngfs_inode));
free:
    brelse(bh);
				return;
}
static int ngfs_fill_super(struct super_block * sb, void *data,int silent){
				struct ngfs_sb_info * nsi;
				struct buffer_head *bh;
				struct ngfs_super_block * ns;
				struct inode *root;
				
				nsi=kmalloc(sizeof(*nsi),GFP_KERNEL);
				if(!nsi)
								return -ENOMEM;
				sb->s_fs_info=nsi;
				memset(nsi,0,sizeof(*nsi));

        /* Read the super block */
				printk(KERN_ALERT "sb_dev = %d, blocksize = %d\n",sb->s_bdev,sb->s_blocksize);
			/*	sb->s_bdev=512; */
			  if (!(bh = sb_bread(sb, NGFS_SUPER_BLOCK))) {
								printk(KERN_ALERT "Cannot read superblock of NGFS \n");
								goto free;
				}
				else
								printk(KERN_ALERT "Read superblock of NGFS in block [%d]\n",NGFS_SUPER_BLOCK);

				ns=(struct ngfs_super_block *)((char *)(bh->b_data));
				nsi->ns=ns;

        /*dump_super_block(ns);*/
        sb->s_magic = ns->s_magic;
				printk(KERN_ALERT "Got magic number as 0x%x\n",sb->s_magic);
				if(sb->s_magic != NGFS_SUPER_MAGIC){
								printk(KERN_ALERT "Not a NGFS File system \n");
								goto free;
				}

				sb->s_op = &ngfs_ops;
				root=iget(sb,NGFS_ROOT_INO);
				sb->s_root = d_alloc_root(root);

				if(!sb->s_root){
								iput(root);
								printk(KERN_ALERT "NGFS:Cannot read root inode \n");
								goto free;
				}

				root->i_fop=&ngfs_dir_ops;
				root->i_op=&ngfs_inode_ops;
				 if (!S_ISDIR(root->i_mode)){
								 dput(sb->s_root);
								 sb->s_root = NULL;
								 printk(KERN_ALERT "NGFS:Error reading root inode \n");
								 goto free;
				 }

				 return 0;

free:
     kfree(nsi);
		 return -EINVAL;
}

static struct super_block * ngfs_get_super(struct file_system_type * fs,int flags, const char * dev_name, void *data){
				return get_sb_bdev(fs, flags, dev_name, data, ngfs_fill_super);
}

/*
void ngfs_kill_super(struct super_block *sb){
				return;
}
*/

static struct file_system_type ngfs_fs_type ={
				name : "ngfs",
				get_sb : ngfs_get_super,
				kill_sb : kill_block_super,
				owner : THIS_MODULE,
				fs_flags : FS_REQUIRES_DEV,
};

static int ngfs_init(void){
				int err=0;
#ifdef DEBUG
				printk(KERN_INFO "Registering Filesystem : \"ngfs\"\n");
#endif
				err=register_filesystem(&ngfs_fs_type); 
				return err;
}

static void ngfs_exit(void){
				int err;
#ifdef DEBUG
				printk(KERN_INFO "Unloading filesystem type : \"ngfs\"\n");
#endif
				if((err=unregister_filesystem(&ngfs_fs_type))!=0) 
				  printk(KERN_INFO "Error Unloading filesystem ngfs : %d\n",err);
}

MODULE_DESCRIPTION("NGFS filesystem");
MODULE_AUTHOR("http://nngfs.pbwiki.com");
MODULE_LICENSE("GPL");
module_init(ngfs_init);
module_exit(ngfs_exit);

