#include<linux/fs.h>

#define NGFS_SB(sb) ((struct ngfs_sb_info *)(sb->s_fs_info))
#define NGFS_DISK_SB(sb) (NGFS_SB(sb)->s_ngfs)
#define NGFS_MAGIC_STR "NgFs V 1.1"
#define NGFS_SUPER_MAGIC 0xdeadbeef
#define NGFS_DEFAULT_BS 512
#define NGFS_SUPER_BLOCK 0 /* Where super block resides */
#define NGFS_INODE_BLOCK 1 /* Block number where inode info is saved  */
#define NGFS_ROOT_BLOCK 2  /* Where root directory info resides */
#define NGFS_MAX_NAME_LEN 25
#define MAGIC_LEN 10
#define NGFS_ROOT_INO 2
#define BLOCK_BMP 4
#define INODE_BMP BLOCK_BMP
#define NGFS_MAX_BLOCKS 10

typedef short unsigned16;

#define FATAL(arg)  \
        fprintf(stderr,arg"%s %d\n",__FILE__,__LINE__); \
				exit(-1);

typedef unsigned int unsigned32;
typedef unsigned char uchar;

struct ngfs_super_block{
				unsigned32 root;
				unsigned32 block_bmp;
				unsigned32 inode_bmp;
				unsigned32 block_size;
				unsigned32 blocks;
				unsigned32 s_magic;
				uchar magic[MAGIC_LEN];
};

struct ngfs_sb_info {
				struct ngfs_super_block * ns;
};

struct ngfs_filesys {
				unsigned int blocksize;
				char * dev_name;
				int fs_dev; /* Descriptor for opened filesystem device */
				struct buffer_head *bh; /* Buffer to hold superblock */
				struct ngfs_super_block * ngfs_ondisk_sb;
				int fs_dirt;
};

struct ngfs_inode {
				unsigned16 ino;
				unsigned16 uid;
				unsigned16 gid;
				unsigned16 mode;
				unsigned16 perm;
				unsigned16 size;
				unsigned16 block_no; /* Block no where data is stored. Remember the restriction, no file is more than blocksize */
				unsigned16 links;
};

struct ngfs_dentry {
				unsigned16 ino;
				/* unsigned16 rec_len; ?? Do we need this ? */
				unsigned16 len;
				unsigned char name[NGFS_MAX_NAME_LEN];
};

void ngfs_create(char * ,unsigned int );
