#include<stdio.h>
#include<fcntl.h>
#include<stdlib.h>
#include<sys/stat.h>
#include<unistd.h>
#include<string.h>
#include "ngfs.h"

#define set_ngfs_bit(bitmap,blockno) bitmap=((bitmap)|(1<<(blockno)))

int bwrite(int dev, int block, int blocksize,char *buff){
				int nw;
				lseek(dev,block*blocksize,SEEK_SET);
				if((nw=write(dev,buff,blocksize))!=blocksize)
								fprintf(stderr,"Could not write completely to the buffer\nWrote %d bytes out of %d\n",nw,blocksize);
				return nw;
}

struct buffer_head * bread(int dev, int block, int blocksize){
				struct buffer_head *bh;
				if((bh=(struct buffer_head *)malloc(sizeof(*bh)))==NULL)
								return NULL;
				bh->b_data=(char *)malloc(blocksize);
				lseek(dev,block*blocksize,SEEK_SET);
				read(dev,bh->b_data,blocksize);
				bh->b_dev=dev;
				bh->b_size=blocksize;
				bh->b_blocknr=block;
				return bh;
}

void create_root_inode(int dev,unsigned32 blockno,unsigned32 inode,unsigned32 blocksize){
				struct buffer_head *bh;
				struct ngfs_inode *root;
				struct ngfs_dentry de;
				bh=bread(dev,blockno,blocksize);
				if(!bh){
								fprintf(stderr,"Could not create root inode\n");
								exit(1);
				}
				root=(struct ngfs_inode *)((char *)bh->b_data+inode*sizeof(*root));

#ifdef DEBUG
        fprintf(stderr,"Writing root inode [%d] info in block [%d]\n",inode,blockno);
#endif
				/*Now write the information of root inode*/
				root->ino=inode;
				root->uid=root->gid=0;
				root->mode|=(S_IFDIR|S_IRUSR|S_IWUSR|S_IXUSR|S_IROTH|S_IXOTH|S_IRGRP|S_IXGRP);
				root->perm=0x0755;
				root->size=blocksize;
				root->block_no=NGFS_ROOT_BLOCK; /* Block no. in which data of root resides */
				root->links=2;
				bwrite(dev,blockno,blocksize,bh->b_data);

        /* Fill the root node with entries of "." and ".." */
				/* Use the prviously allocated b_data to fill root node info */
				memset(bh->b_data,0,blocksize);
				memset(&de,0,sizeof(de));
				de.ino=inode;
				de.len=1;
				strcpy(de.name,".");
				memcpy(bh->b_data,&de,sizeof(de));
				memset(&de,0,sizeof(de));
				de.ino=inode;
				de.len=2;
				strcpy(de.name,"..");
				memcpy(bh->b_data+sizeof(de),&de,sizeof(de));
				bwrite(dev,NGFS_ROOT_BLOCK,blocksize,bh->b_data);
				return;
}

void ngfs_create(char * filename,unsigned int bs){
				struct ngfs_filesys *fs;
				struct ngfs_super_block *sb;

        if(bs<=0 || bs>4096){
#ifdef DEBUG
								fprintf(stderr,"Either invalid blocksize or too big for NGFS !!..setting to default [%d]\n",NGFS_DEFAULT_BS);
#endif
								bs=NGFS_DEFAULT_BS;
				}
				fs=(struct ngfs_filesys *)malloc(sizeof(*fs));
				fs->fs_dev=open(filename,O_RDWR);
				if(fs->fs_dev==-1){
								perror("Cannot open partition ");
								goto free;
				}

				fs->blocksize=bs; /* Set the blocksize */
#ifdef DEBUG
				fprintf(stderr,"Writing super block info in block [%d]\n",NGFS_SUPER_BLOCK);
#endif
				fs->bh=bread(fs->fs_dev,NGFS_SUPER_BLOCK,bs); /* Get buffer of superblock */
				if(!fs->bh){
#ifdef DEBUG
								fprintf(stderr,"Could not get buffer for superblock \n");
#endif
								goto free;
				}

        fs->ngfs_ondisk_sb = (struct ngfs_super_block *)(fs->bh->b_data);
				memset(fs->ngfs_ondisk_sb,0,sizeof(struct ngfs_super_block));

				sb=fs->ngfs_ondisk_sb;
				sb->root=NGFS_ROOT_INO;
				set_ngfs_bit(sb->block_bmp,NGFS_SUPER_BLOCK);
				set_ngfs_bit(sb->block_bmp,NGFS_ROOT_BLOCK);
				set_ngfs_bit(sb->inode_bmp,NGFS_ROOT_INO);
				sb->blocks=NGFS_MAX_BLOCKS;
				sb->s_magic=NGFS_SUPER_MAGIC;
				strcpy(sb->magic,NGFS_MAGIC_STR);
				bwrite(fs->fs_dev,NGFS_SUPER_BLOCK,bs,fs->bh->b_data);

				create_root_inode(fs->fs_dev,NGFS_INODE_BLOCK,NGFS_ROOT_INO,fs->blocksize);

				fprintf(stderr,"NGFS filesystem created successfully on \"%s\"\n",filename);
				return;
free:
     free(fs);
		 return;
}
